//
//  CoreDataMapsViewController.m
//  CoreDataMaps
//
//  Created by Patrick Hughes on 2/10/10.
//  Copyright Fluffy Bunny Software 2010. All rights reserved.
//

#import "CoreDataMapsViewController.h"
#import "LocalSearchResult.h"
#import <QuartzCore/QuartzCore.h>

@interface CoreDataMapsViewController (private)
-(void)reloadData;
-(void)dismissPanel:(id)sender;
-(void)showPanel:(BOOL)visible animated:(BOOL)animated;
-(NSArray*)sortDescriptors;
-(NSPredicate*)predicate;
-(NSPredicate*)predicateTemplate;
- (NSString *)applicationDocumentsDirectory;
@end

@implementation CoreDataMapsViewController
@synthesize fetchedResultsController, managedObjectContext;

-(IBAction)addLocation:(id)sender {
	[self showPanel:YES animated:YES];
	
	// Bring in the keyboard.
	[nameField becomeFirstResponder];
}


- (BOOL)textFieldShouldReturn:(UITextField *)textField {
	// create a new location object
	LocalSearchResult *l = [LocalSearchResult insertInManagedObjectContext:[self managedObjectContext]];

	// and set it's attributes.
	// These are all mandatory (as set in the data model designer) saving without setting will
	// cause an exception.
	l.latitude = [NSNumber numberWithDouble:mapView.region.center.latitude];
	l.longitude = [NSNumber numberWithDouble:mapView.region.center.longitude];
	l.name = nameField.text;
	
	// I find it looks better if we let the pin land before reverting back.
	[self performSelector:@selector(dismissPanel:) withObject:nil afterDelay:0.4f];
	
	return YES;
}

-(void)dismissPanel:(id)sender {
	[self showPanel:NO animated:YES];
	
	// move out the keyboard
	[nameField resignFirstResponder];
}

-(void)showPanel:(BOOL)visible animated:(BOOL)animated {
	CGRect mapFrame = mapView.frame;
	CGRect formFrame = formOverlay.frame;
	MKCoordinateRegion r = mapView.region;
	CGFloat alpha;
	
	if (visible) {
		// zoom in?
		//r.span = MKCoordinateSpanMake(0.001f, 0.001f);
		// set the visiblity of the crosshairs
		alpha = 1.0f;
		
		// move the map up
		mapFrame.origin.y = -95;
		
		// move the form down.
		formFrame.origin.y = -10;
		
		// swap out the button
		UIBarButtonItem *button = [[[UIBarButtonItem alloc] initWithBarButtonSystemItem:UIBarButtonSystemItemCancel target:self action:@selector(dismissPanel:)] autorelease];
		[self.navigationItem setRightBarButtonItem:button animated:YES];
	}
	else {
		// set the visiblity of the crosshairs
		alpha = 0.0f;		
		// move the map down
		mapFrame.origin.y =  0;

		// move the form up.
		formFrame.origin.y = -55;
		
		// swap the add button back in
		UIBarButtonItem *button = [[[UIBarButtonItem alloc] initWithBarButtonSystemItem:UIBarButtonSystemItemAdd target:self action:@selector(addLocation:)] autorelease];
		[self.navigationItem setRightBarButtonItem:button animated:YES];
	}
	
	if (animated) {
		[UIView beginAnimations:@"animatePanel" context:nil];
	}
	formOverlay.frame = formFrame;
	mapView.frame = mapFrame;
	mapView.region = r;
	mapOverlay.alpha = alpha;
	mapOverlay.frame = mapFrame;
	
	if(animated) {
		[UIView commitAnimations];
	}
}


-(void)reloadData {
	NSError *error = nil;
	
	if (![[self fetchedResultsController] performFetch:&error]) {
		
		// I did a bad thing and copied and pasted this code.
		if([error code] == 134100) {
			NSLog(@"Deleting incompatible store");
			
			NSFileManager *fm = [NSFileManager defaultManager];
			NSError *fileError = nil;
			if([fm removeItemAtPath:[[self applicationDocumentsDirectory] stringByAppendingPathComponent: @"CoreDataMaps.sqlite"] error:&fileError] == FALSE) {
				NSLog(@"failed to delete store: %@", fileError);
			}
		}
		/*
		 Replace this implementation with code to handle the error appropriately.
		 
		 abort() causes the application to generate a crash log and terminate. 
		 You should not use this function in a shipping application, although it may be useful during development. 
		 If it is not possible to recover from the error, 
		 display an alert panel that instructs the user to quit the application by pressing the Home button.
		 */
		
		NSLog(@"Unresolved error %@, %@", error, [error userInfo]);
		abort();
	}
	NSArray *fetchedObjects = [[self fetchedResultsController] fetchedObjects];
	NSLog(@"fetched %d items", [fetchedObjects count]);
	
	// For each item in fetchedObjects, if the item hasn't been added to the map yet 
	// the delegate method mapView:viewForAnnotation: is called. 
	[mapView addAnnotations:fetchedObjects];
}

-(void)saveData {
	NSLog(@"saving data");
	NSError *error = nil;
	if ([[self managedObjectContext] hasChanges] && ![[self managedObjectContext] save:&error]) {
		/*
		 Replace this implementation with code to handle the error appropriately.
		 
		 abort() causes the application to generate a crash log and terminate. You should not use this function in a shipping application, although it may be useful during development. If it is not possible to recover from the error, display an alert panel that instructs the user to quit the application by pressing the Home button.
		 */
		NSLog(@"Unresolved error %@, %@", error, [error userInfo]);
		abort();
	} 	
}

#pragma mark MKMapViewDelegate methods
- (void)mapView:(MKMapView *)mapView_ regionDidChangeAnimated:(BOOL)animated {
	// When the size or center of the map changes reset the predicate to only
	// load locations on the map. This doesn't remove annotations located off of the
	// map. We would have to do that manually.
	NSLog(@"region did change");
	[[[self fetchedResultsController] fetchRequest] setPredicate:[self predicate]];
	[self reloadData];
}

- (MKAnnotationView *)mapView:(MKMapView *)mapView_ viewForAnnotation:(id <MKAnnotation>)annotation {
	MKPinAnnotationView *view = (MKPinAnnotationView*)[mapView_ dequeueReusableAnnotationViewWithIdentifier:@"mapPin"];
	if(view == nil) {
		view = [[[MKPinAnnotationView alloc] initWithAnnotation:annotation reuseIdentifier:@"mapPin"] autorelease];
		// The map view can optionally show the user's location. I like to make it look different.
		if([annotation isKindOfClass:[MKUserLocation class]]) {
			[(MKUserLocation*)annotation setTitle:@"Here"];
			view.pinColor = MKPinAnnotationColorGreen;
		}
		else {
			view.pinColor = MKPinAnnotationColorRed;
			// Here is where we would add accessory controls.
		}
		view.canShowCallout = YES;
		view.animatesDrop = YES;
	}
	view.annotation = annotation;
	return view;
}

- (void)mapView:(MKMapView *)mapView_ annotationView:(MKAnnotationView *)view calloutAccessoryControlTapped:(UIControl *)control {
	if (view.leftCalloutAccessoryView == control) {
		// We haven't set any accessory controls, but if we had we would deal with the resulting actions here.
	}
	else if (view.rightCalloutAccessoryView == control) {
		
	}
	else {
		NSLog(@"something else tapped");
	}

}

#pragma mark -
#pragma mark Fetched results controller

- (NSFetchedResultsController *)fetchedResultsController {    
	// I love this pattern for accessors, we need it, but we can release it
	// if memory gets too low, and (as long as we're using the accessor) it
	// will be recreated the next time we need it. Lazy loading at it's best.
    if (fetchedResultsController != nil) {
        return fetchedResultsController;
    }

	// Create the fetch request for the entity.
	NSFetchRequest *fetchRequest = [[NSFetchRequest alloc] init];
	// Edit the entity name as appropriate.
	NSEntityDescription *entity = [NSEntityDescription entityForName:@"LocalSearchResult" inManagedObjectContext:[self managedObjectContext]];
	
	[fetchRequest setEntity:entity];

	// Set the batch size to a suitable number.
	[fetchRequest setFetchBatchSize:20];
	
	// I like to make separate methods for the predicate and sort descriptors
	// This lets me reuse this method pretty much as is.
	[fetchRequest setPredicate:[self predicate]];	
	[fetchRequest setSortDescriptors:[self sortDescriptors]];
	
	NSFetchedResultsController *aFetchedResultsController = [[NSFetchedResultsController alloc] initWithFetchRequest:fetchRequest 
																								managedObjectContext:managedObjectContext
																								  sectionNameKeyPath:nil
																										   cacheName:@"maps"];
    aFetchedResultsController.delegate = self;
	self.fetchedResultsController = aFetchedResultsController;
	
	[aFetchedResultsController release];
	[fetchRequest release];
	
	return fetchedResultsController;
}    

-(NSArray*)sortDescriptors {
	// In this case we probably don't need to worry about sorting.
	// We could create an instance method that computed distance and sort with that,
	// but not today.
	NSSortDescriptor *sortDescriptor = [[NSSortDescriptor alloc] initWithKey:@"name" ascending:YES];
	NSArray *sortDescriptors = [NSArray arrayWithObject:sortDescriptor];
	[sortDescriptor release];

	return sortDescriptors;
}

-(NSPredicate*)predicate {
	// request only those items in the span of the map
	CLLocationDegrees lat = mapView.region.center.latitude;
	CLLocationDegrees lon = mapView.region.center.longitude;
	CLLocationDegrees latDelta = mapView.region.span.latitudeDelta / 2.0f;
	CLLocationDegrees lonDelta = mapView.region.span.longitudeDelta / 2.0f;
	
	NSMutableDictionary *dict = [NSMutableDictionary dictionaryWithCapacity:4];
	[dict setObject: [NSNumber numberWithDouble:lat - latDelta] forKey:@"MIN_LAT"];
	[dict setObject: [NSNumber numberWithDouble:lat + latDelta] forKey:@"MAX_LAT"];
	[dict setObject: [NSNumber numberWithDouble:lon - lonDelta] forKey:@"MIN_LON"];
	[dict setObject: [NSNumber numberWithDouble:lon + lonDelta] forKey:@"MAX_LON"];

	// fill the predicate template with our computed values
	// using predicate templates is covered in the Predicate Programming Guide 
	// file:///Developer/Platforms/iPhoneOS.platform/Developer/Documentation/DocSets/com.apple.adc.documentation.AppleiPhone3_1.iPhoneLibrary.docset/Contents/Resources/Documents/documentation/Cocoa/Conceptual/Predicates/Articles/pCreating.html
	NSPredicate *p = [[self predicateTemplate] predicateWithSubstitutionVariables:dict];
	NSLog(@"%@", p);
	return p;
}

-(NSPredicate*)predicateTemplate {
	if (predicateTemplate != nil) {
		return predicateTemplate;
	}
	// we need our predicate template to search with
	predicateTemplate = [[NSPredicate predicateWithFormat:@"((latitude > $MIN_LAT AND latitude < $MAX_LAT) AND (longitude > $MIN_LON AND longitude < $MAX_LON))"] retain];
	return predicateTemplate;
}


#pragma mark NSFetchedResultsControllerDelegate method 
- (void)controller:(NSFetchedResultsController *)controller didChangeObject:(id)anObject
    atIndexPath:(NSIndexPath *)indexPath forChangeType:(NSFetchedResultsChangeType)type
    newIndexPath:(NSIndexPath *)newIndexPath {
 
	id <MKAnnotation> object = (id <MKAnnotation>)anObject;
	
    switch(type) {
 
        case NSFetchedResultsChangeInsert:
			// add the new object to the map
			[mapView addAnnotation:object];

            break;
			
			// We haven't introduced anything that would make this happen.		
        case NSFetchedResultsChangeDelete:
            [mapView removeAnnotation:object];
            break;
			
			// We haven't introduced anything that would make this happen.		
        case NSFetchedResultsChangeUpdate:
			NSLog(@"annotation changed");
            break;
    }
}
 
-(void)viewWillAppear:(BOOL)animated {
	CLLocationCoordinate2D c;
	c.latitude = 43.028368;
	c.longitude = -76.038923;
	
	MKCoordinateSpan s;
	s.latitudeDelta = 0.02;
	s.longitudeDelta = 0.02;
	
	MKCoordinateRegion r;
	r.center = c;
	r.span = s;
	
	[mapView setRegion:r animated:YES];
}



// Implement viewDidLoad to do additional setup after loading the view, typically from a nib.
- (void)viewDidLoad {
	// The background color for the form is set in the nib. %80 opaque black.
	
	// Round the corners of the view. I used to write dozens of lines of code to do this.
	CALayer *l = [formOverlay layer];
	l.cornerRadius = 10.0f;
	
	// Make a pretty border, because we can.
	l.borderColor = [[UIColor lightGrayColor] CGColor];
	l.borderWidth = 2.0f;
	
	// dismissing the panel will configure all of the panel elements
	[self showPanel:NO animated:NO];
    [super viewDidLoad];
}


- (void)didReceiveMemoryWarning {
    [super didReceiveMemoryWarning];
	
	// Release anything that can be recreated easily.
	[fetchedResultsController release];
	fetchedResultsController = nil;
	[predicateTemplate release];
	predicateTemplate = nil;
}

- (void)dealloc {
	[mapView release];
	[mapOverlay release];
	[formOverlay release];
	[nameField release];
	
	[fetchedResultsController release];
	[predicateTemplate release];
	
    [super dealloc];
}

@end
